use super::{Datum, DatumList, DatumListRef, DatumVisitor};

impl DatumList for () {
    type Ref<'a> = &'a ();
    fn accept(&self, _: &mut impl DatumVisitor) {}
}

impl<'l> DatumListRef for &'l () {
    fn accept(&self, _: &mut impl DatumVisitor) {}
}

impl<T: Datum> DatumList for T {
    type Ref<'a> = &'a T where Self: 'a;

    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(self);
    }
}

impl<'l, T: Datum> DatumListRef for &'l T {
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(self);
    }
}

impl<T0: Datum> DatumList for (T0,) {
    type Ref<'a> = (&'a T0,) where Self: 'a;

    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
    }
}

impl<'a, T0: Datum> DatumListRef for (&'a T0,) {
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
    }
}

macro_rules! datum_list {
    ($($ty:ident : $n:tt),+) => {
        impl<$($ty: Datum),*> DatumList for ($($ty),*) {
            type Ref<'a> = ($(&'a $ty),*) where Self: 'a;
            fn accept(&self, visitor: &mut impl DatumVisitor) {
                $(visitor.visit(&self. $n));*
            }
        }

        impl<'l, $($ty: Datum),*> DatumListRef for ($(&'l $ty),*) {
            fn accept(&self, visitor: &mut impl DatumVisitor) {
                $(visitor.visit(&self. $n));*
            }
        }
    }
}

datum_list!(T0: 0, T1: 1);
datum_list!(T0: 0, T1: 1, T2: 2);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5, T6: 6);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5, T6: 6, T7: 7);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5, T6: 6, T7: 7, T8: 8);

/*
impl<T0: Datum, T1: Datum> DatumList for (T0, T1) {
    type Ref<'a> = (&'a T0, &'a T1) where Self: 'a;
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
    }
}

impl<'a, T0: Datum, T1: Datum> DatumListRef for (&'a T0, &'a T1) {
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
    }
}

impl<T0: Datum, T1: Datum, T2: Datum> DatumList for (T0, T1, T2) {
    type Ref<'a> = (&'a T0, &'a T1, &'a T2) where Self: 'a;
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
        visitor.visit(&self.2);
    }
}

impl<'a, T0: Datum, T1: Datum, T2: Datum> DatumListRef for (&'a T0, &'a T1, &'a T2) {
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
        visitor.visit(&self.2);
    }
}

impl<T0: Datum, T1: Datum, T2: Datum, T3: Datum> DatumList for (T0, T1, T2, T3) {
    type Ref<'a> = (&'a T0, &'a T1, &'a T2, &'a T3) where Self: 'a;
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
        visitor.visit(&self.2);
        visitor.visit(&self.3);
    }
}

impl<T0: Datum, T1: Datum, T2: Datum, T3: Datum, T4: Datum> DatumList for (T0, T1, T2, T3, T4) {
    type Ref<'a> = (&'a T0, &'a T1, &'a T2, &'a T3, &'a T4);
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
        visitor.visit(&self.2);
        visitor.visit(&self.3);
        visitor.visit(&self.4);
    }
}

impl<T0: Datum, T1: Datum, T2: Datum, T3: Datum, T4: Datum, T5: Datum> DatumList
    for (T0, T1, T2, T3, T4, T5)
{
    type Ref<'a> = (&'a T0, &'a T1, &'a T2, &'a T3, &'a T4, &'a T5);
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
        visitor.visit(&self.1);
        visitor.visit(&self.2);
        visitor.visit(&self.3);
        visitor.visit(&self.4);
        visitor.visit(&self.5);
    }
}
*/
